/* Test of kd-tree building code.
 *
 * Daniel Lowd, 5/9/03
 */

#include <stdio.h>
#include <stdlib.h>

/* ****** BEGIN DUAL MCL STUFF ****** */

#define MAX_X 3000
#define MAX_Y 3000
#define MIN_X -3000
#define MIN_Y -3000
#define X_VAR 0
#define Y_VAR 1
#define A_VAR 2

struct kd_data
{
    double* vars;
};

// A single node in a kd-tree
struct s_kd_node
{
    int split_var;
    double split_val;
    struct s_kd_node* left;
    struct s_kd_node* right;

    // Holds final child in leaf nodes
    struct kd_data data;
};

struct s_kd_rect
{
    double min_x;
    double max_x;
    double min_y;
    double max_y;
    double min_a;
    double max_a;
};

typedef struct s_kd_node kd_node;
typedef struct s_kd_rect kd_rect;

int kd_var = 0;

int kd_comp(const void* first, const void* second)
{
    double first_val = ((kd_data*)first)->vars[kd_var];
    double second_val = ((kd_data*)second)->vars[kd_var];

    if (first_val < second_val)
        return -1;
    else if (first_val > second_val)
        return 1;
    else
        return 0;
}

kd_node* build_tree(int first_var, int max_var, int num_data, 
        struct kd_data* data)
{
    kd_node* ret = (kd_node*)malloc(sizeof(kd_node));
    ret->split_var = first_var;

    // Degenerate case -- 1 data point left.  End recursion.
    if (num_data == 1)
    {
        ret->split_val = data[0].vars[first_var];
        ret->left      = NULL;
        ret->right     = NULL;
        ret->data      = data[0];
        return ret;
    }
    else
    {
        // Prepare to recurse with half the data
        int next_var = (first_var + 1) % (max_var+1);
        int num_left = num_data/2;
        int num_right = num_data - num_left;

        // Sort by specified variable (required to find median)
        // We use a global var to tell the comparison routine (kd_comp) which
        // of the k variables we're using for comparison.
        kd_var = first_var;
        qsort(data, num_data, sizeof(kd_data), kd_comp);

        for (int i = 0; i < num_data; i++)
            printf("%f ", data[i].vars[kd_var]);
        putchar('\n');

        ret->split_val = data[num_left].vars[first_var];
        printf("split_val: %f\n", ret->split_val);
        ret->left = build_tree(next_var, max_var, num_left, data);
        ret->right = build_tree(next_var, max_var, num_right, data + num_left);
        return ret;
    }
}

void print_tabs(int numtabs)
{
    while (numtabs--) {
        putchar('\t');
    }
}

void print_tree(kd_node* curr_node, int curr_level, int max_var)
{
    if (curr_node->left == NULL)
    {
        print_tabs(curr_level);
        putchar('(');
        for (int i = 0; i <= max_var; i++) { 
            if (i != 0)
                putchar(',');
            printf("%f", curr_node->data.vars[i]);
        }
        printf(")\n");
        return;
    }

    print_tabs(curr_level);
    printf("var: %d; split: %f\n", curr_node->split_var, curr_node->split_val);
    print_tabs(curr_level);
    printf("LEFT CHILD\n");
    print_tree(curr_node->left, curr_level+1, max_var);
    print_tabs(curr_level);
    printf("RIGHT CHILD\n");
    print_tree(curr_node->right, curr_level+1, max_var);
}

//////////////// TESTING //////////////////

int main()
{
    double testdoubles[6][3] = {{1.01, 2.02, 3.03}, {2.11, 4.12, 1.13}, 
        {5.21, 4.22, 4.23}, {0.31, 3.32, 1.33}, {5.41, 2.42, 3.43}, {2.51, 2.52, 2.53}};
    kd_data testdata[6] = {
        {testdoubles[0]},
        {testdoubles[1]},
        {testdoubles[2]},
        {testdoubles[3]},
        {testdoubles[4]}, 
        {testdoubles[5]}};

    kd_node* root = build_tree(0, 2, 6, testdata);
    print_tree(root, 0, 2);
    return 0;
}
